/*
Module : JSON++.h
Purpose: Defines the interface for a set of classes to support JSON parsing and creation. The class is based
         in part on the SimpleJSON C++ class of Mike Anchor available at https://github.com/MJPA/SimpleJSON
History: PJN / 25-11-2017 1. Initial public release.
         PJN / 10-01-2018 1. Updated copyright details.
                          2. Added support for JSON Pointer (https://tools.ietf.org/html/rfc6901) support to the code
                          via a new CValue::JSONPointer method.
                          3. Made a number of methods of the CValue class static as they did not need to be non-static.
         PJN / 29-05-2018 1. Fixed a number of C++ core guidelines compiler warnings.
         PJN / 30-06-2018 1. Reworked the implementation to use std::variant. This does mean that the code now requires
                          at least the /std:c++17 compiler setting. This has resulted in the module shrinking by c. 160 
                          lines. Testing shows the performance remains the same as before.
         PJN / 31-07-2018 1. Reworked code to use Windows API calls MultiByteToWideChar API  for the UTF8 to UTF16 
                          conversion rather than the deprecated codecvt header.
         PJN / 02-09-2018 1. Fixed a number of compiler warnings when using VS 2017 15.8.2.
         PJN / 22-04-2019 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2019.
         PJN / 03-08-2019 1. Added comprehensive SAL annotations to all the code.
         PJN / 29-01-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 23-05-2020 1. Added const versions of the CValue::Child methods.
                          2. Added [[nodiscard]] to a number of CValue methods.
         PJN / 02-11-2021 1. Updated copyright details.
                          2. Updated compiler settings to more modern defaults.
         PJN / 29-01-2022 1. Updated copyright details.
                          2. Fixed more static code analysis warnings in Visual Studio 2022.
                          3. Updated the code to use C++ uniform initialization for all variable declarations.
         PJN / 22-04-2022 1. Fixed a bug in CValue::operator=(const wchar_t*) where the m_Data member 
                          variable would get initialized incorrectly to a bool on VS 2022. Thanks to "jls"
                          for reporting this issue.
         PJN / 12-05-2023 1. Updated copyright details
                          2. Updated modules to indicate that it needs to be compiled using /std:c++17. Thanks to Martin Richter
                          for reporting this issue.
         PJN / 09-07-2023 1. Updated the class to provide support for JSON5 (https://spec.json5.org/) parsing and encoding. 
                          The JSON5 implementation in JSON++ follows the guidelines which SQLite3 used 
                          (https://www.sqlite.org/json1.html#json5) when it was updated to support JSON5. To support parsing of 
                          JSON5, you can pass the JSONPP::PARSE_FLAGS::PARSE_JSON5 value as the third parameter to the 
                          CValue::Parse method. If you pass a zero instead for this parameter then the method will revert to 
                          strict JSON parsing. To support JSON5 encoding you can pass the JSONPP::ENCODE_FLAGS::ENCODE_JSON5 
                          value as the first parameter to the CValue::Encode method. Please see the example code in main.cpp 
                          for further examples.

Copyright (c) 2017 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


///////////////////////// Macros / Defines ////////////////////////////////////

#pragma once

#if _MSVC_LANG < 201703
#error JSON++ requires a minimum C++ language standard of /std:c++17
#endif //#if _MSVC_LANG < 201703

#ifndef __JSONPP_H__
#define __JSONPP_H__


///////////////////////// Includes ////////////////////////////////////////////

#include <windows.h>
#include <string>
#include <vector>
#include <map>
#include <cassert>
#include <sstream>
#include <cmath>
#include <algorithm>
#include <locale>
#include <type_traits>
#include <variant>
#include <limits>


///////////////////////// Implementation  /////////////////////////////////////

//Namespaces
namespace JSONPP
{


//Forward references
class CValue;

//Typedefs
using Object = std::map<std::wstring, CValue>;
using Array = std::vector<CValue>;

//Enums
enum class PARSE_EXCEPTION_REASON
{
  ENCODED_UNICODE_CHARACTER_TOO_SHORT = 0,
  UNEXPECTED_CHARACTER_IN_ENCODED_UNICODE_CHARACTER = 1,
  UNEXPECTED_CHARACTER_IN_STRING = 2,
  UNEXPECTED_CHARACTER_IN_NUMBER = 3,
  UNEXPECTED_CHARACTER_IN_ARRAY = 4,
  UNEXPECTED_CHARACTER_IN_OBJECT = 5,
  UNEXPECTED_CHARACTER_IN_JSON = 6,
  UNEXPECTED_CHARACTER_IN_JSON_POINTER = 7, //Only thrown by JSONPointer method
  COULD_NOT_FIND_CHILD_IN_OBJECT = 8, //Only thrown by JSONPointer method
  COULD_NOT_FIND_CHILD_IN_ARRAY = 9, //Only thrown by JSONPointer method
  CURRENT_ITEM_IS_NOT_OBJECT_OR_ARRAY = 10, //Only thrown by JSONPointer method
  UNEXPECTED_CHARACTER_IN_MULTI_LINE_COMMENT = 11, //Only thrown when parsing JSON5
  UNEXPECTED_CHARACTER_IN_JSON5_IDENTIFIER = 12 //Only thrown when parsing JSON5
};

enum PARSE_FLAGS
{
  PARSE_JSON5 = 0x0001
};

enum ENCODE_FLAGS
{
  NONE = 0x0000,
  TABS = 0x0001,
  ONE_SPACE = 0x0002,
  TWO_SPACES = 0x0004,
  THREE_SPACES = 0x0008,
  FOUR_SPACES = 0x0010,
  WHITESPACE_BETWEEN_OBJECT_NAME_AND_VALUES = 0x0020,
  WHITESPACE_AFTER_VALUE = 0x0040,
  WHITESPACE_BEFORE_VALUE = 0x0080,
  INDENT_SUB_OBJECTS = 0x0100,
  INDENT_SUB_ARRAYS = 0x0200,
  PRETTY_PRINT_ONE_SPACE = 0x0302, //A combination of INDENT_SUB_ARRAYS, INDENT_SUB_OBJECTS & ONE_SPACE
  ENCODE_NON_ASCII_CHARACTERS = 0x0400,
  ENCODE_ALL_UNICODE_CHARACTERS = 0x0800,
  ENCODE_JSON5 = 0x1000
};

enum class TYPE
{
  NULL_VALUE = 0,
  STRING = 1,
  NUMBER = 2,
  OBJECT = 3,
  ARRAY = 4,
  BOOLEAN = 5
};

//Classes
class CParseException : public std::exception
{
public:
//Constructors / Destructors
  CParseException() = delete;
  CParseException(_In_ PARSE_EXCEPTION_REASON reason, _In_ std::size_t nIndex) noexcept : m_Reason{reason},
                                                                                          m_nIndex{nIndex}
  {
  }
  CParseException(const CParseException&) = default;
  CParseException(CParseException&&) = default;
#pragma warning(suppress: 26443)
  ~CParseException() override = default;

//Methods
  CParseException& operator=(const CParseException&) = default;
  CParseException& operator=(CParseException&&) = default;

//Member variables
  PARSE_EXCEPTION_REASON m_Reason; //The reason for the parsing error
  std::size_t m_nIndex; //The 0 based index from the beginning of the string where the parsing error occurred
#pragma warning(suppress: 26495)
};

class CValue
{
public:
//Constructors / Destructors
  CValue() noexcept : m_Type{TYPE::NULL_VALUE}
  {
  }

  ~CValue() = default;
  CValue(const CValue&) = default;
  CValue(CValue&& value) = default;

#pragma warning(suppress: 26440)
  CValue(_In_ const std::wstring& sString) : m_Data{sString},
                                             m_Type{TYPE::STRING}
  {
  }

#pragma warning(suppress: 26440)
  CValue(_In_ std::wstring&& sString) : m_Data{sString},
                                        m_Type{TYPE::STRING}
  {
  }

#pragma warning(suppress: 26440)
  CValue(_In_z_ const wchar_t* pszString) : m_Data{std::wstring(pszString)},
                                            m_Type{TYPE::STRING}
  {
  }

#pragma warning(suppress: 26440)
  CValue(_In_ double fNumber) : m_Data{fNumber},
                                m_Type{TYPE::NUMBER}
  {
  }

#pragma warning(suppress: 26440)
  CValue(_In_ long nNumber) : m_Data{static_cast<double>(nNumber)},
                              m_Type{TYPE::NUMBER}
  {
  }

#pragma warning(suppress: 26440)
  CValue(_In_ const Object& object) : m_Data{object},
                                      m_Type{TYPE::OBJECT}
  {
  }

  CValue(_In_ Object&& object) noexcept : m_Data{object},
                                          m_Type{TYPE::OBJECT}
  {
  }

#pragma warning(suppress: 26440)
  CValue(_In_ const Array& array) : m_Data{array},
                                    m_Type{TYPE::ARRAY}
  {
  }

#pragma warning(suppress: 26440)
  CValue(_In_ Array&& array) : m_Data{array},
                               m_Type{TYPE::ARRAY}
  {
  }

#pragma warning(suppress: 26440)
  CValue(_In_ bool bValue) : m_Data{bValue},
                             m_Type{TYPE::BOOLEAN}
  {
  }

//Methods
  void SetNull() noexcept
  {
    m_Type = TYPE::NULL_VALUE;
  }

  CValue& operator=(_In_ const std::wstring& sString)
  {
    m_Data = sString;
    m_Type = TYPE::STRING;
    return *this;
  }

#pragma warning(suppress: 26440)
  CValue& operator=(_In_z_ const wchar_t* pszString)
  {
    m_Data = std::wstring(pszString);
    m_Type = TYPE::STRING;
    return *this;
  }

#pragma warning(suppress: 26440)
  CValue& operator=(_In_ std::wstring&& sString)
  {
    m_Data = std::move(sString);
    m_Type = TYPE::STRING;
    return *this;
  }

#pragma warning(suppress: 26440)
  CValue& operator=(_In_ double fNumber)
  {
    m_Data = fNumber;
    m_Type = TYPE::NUMBER;
    return *this;
  }

#pragma warning(suppress: 26440)
  CValue& operator=(_In_ long nNumber)
  {
    m_Data = static_cast<double>(nNumber);
    m_Type = TYPE::NUMBER;
    return *this;
  }

  CValue& operator=(_In_ const Object& object)
  {
    m_Data = object;
    m_Type = TYPE::OBJECT;
    return *this;
  }

  CValue& operator=(_In_ Object&& object)
  {
    m_Data = std::move(object);
    m_Type = TYPE::OBJECT;
    return *this;
  }

  CValue& operator=(_In_ const Array& array)
  {
    m_Data = array;
    m_Type = TYPE::ARRAY;
    return *this;
  }

#pragma warning(suppress: 26440)
  CValue& operator=(_In_ Array&& array)
  {
    m_Data = std::move(array);
    m_Type = TYPE::ARRAY;
    return *this;
  }

#pragma warning(suppress: 26440)
  CValue& operator=(_In_ bool bValue)
  {
    m_Data = bValue;
    m_Type = TYPE::BOOLEAN;
    return *this;
  }

  CValue& operator=(_In_ const CValue& value)
  {
    switch (value.m_Type)
    {
      case TYPE::NULL_VALUE:
      {
        //Nothing to do
        break;
      }
      case TYPE::STRING:
      {
        m_Data = value.AsString();
        break;
      }
      case TYPE::NUMBER:
      {
        m_Data = value.AsNumber();
        break;
      }
      case TYPE::OBJECT:
      {
        m_Data = value.AsObject();
        break;
      }
      case TYPE::ARRAY:
      {
        m_Data = value.AsArray();
        break;
      }
      case TYPE::BOOLEAN:
      {
        m_Data = value.AsBoolean();
        break;
      }
      default:
      {
      #ifdef _DEBUG
        assert(false);
      #endif //#ifdef _DEBUG
        break;
      }
    }
    m_Type = value.m_Type;
    return *this;
  }

#pragma warning(suppress: 26439)
  CValue& operator=(_In_ CValue&& value)
  {
    switch (value.m_Type)
    {
      case TYPE::NULL_VALUE:
      {
        break;
      }
      case TYPE::STRING:
      {
        m_Data = std::move(value.AsString());
        break;
      }
      case TYPE::NUMBER:
      {
        m_Data = value.AsNumber();
        break;
      }
      case TYPE::OBJECT:
      {
        m_Data = std::move(value.AsObject());
        break;
      }
      case TYPE::ARRAY:
      {
        m_Data = std::move(value.AsArray());
        break;
      }
      case TYPE::BOOLEAN:
      {
        m_Data = value.AsBoolean();
        break;
      }
      default:
      {
      #ifdef _DEBUG
        assert(false);
      #endif //#ifdef _DEBUG
        break;
      }
    }
    m_Type = value.m_Type;
    return *this;
  }

  [[nodiscard]] inline TYPE Type() const noexcept
  {
    return m_Type;
  }

  [[nodiscard]] inline bool IsNull() const noexcept
  {
    return m_Type == TYPE::NULL_VALUE;
  }

  [[nodiscard]] inline bool IsString() const noexcept
  {
    return m_Type == TYPE::STRING;
  }

  [[nodiscard]] inline bool IsNumber() const noexcept
  {
    return m_Type == TYPE::NUMBER;
  }

  [[nodiscard]] inline bool IsArray() const noexcept
  {
    return m_Type == TYPE::ARRAY;
  }

  [[nodiscard]] inline bool IsObject() const noexcept
  {
    return m_Type == TYPE::OBJECT;
  }

  [[nodiscard]] inline bool IsBoolean() const noexcept
  {
    return m_Type == TYPE::BOOLEAN;
  }

  [[nodiscard]] inline const std::wstring AsString() const
  {
  //Validate our parameters
  #ifdef _DEBUG
    assert(m_Type == TYPE::STRING);
  #endif //#ifdef _DEBUG

    return std::get<std::wstring>(m_Data);
  }

#pragma warning(suppress: 26440)
  [[nodiscard]] inline std::wstring& AsString()
  {
  //Validate our parameters
  #ifdef _DEBUG
    assert(m_Type == TYPE::STRING);
  #endif //#ifdef _DEBUG

    return std::get<std::wstring>(m_Data);
  }

#pragma warning(suppress: 26440)
  [[nodiscard]] inline double AsNumber() const
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_Type == TYPE::NUMBER);
  #endif //#ifdef _DEBUG

    return std::get<double>(m_Data);
  }

#pragma warning(suppress: 26440)
  [[nodiscard]] inline double& AsNumber()
  {
    //Validate our parameters
  #ifdef _DEBUG
assert(m_Type == TYPE::NUMBER);
#endif //#ifdef _DEBUG

    return std::get<double>(m_Data);
  }

#pragma warning(suppress: 26440)
  [[nodiscard]] inline const Array& AsArray() const
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_Type == TYPE::ARRAY);
#endif //#ifdef _DEBUG

    return std::get<Array>(m_Data);
  }

#pragma warning(suppress: 26440)
  [[nodiscard]] inline Array& AsArray()
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_Type == TYPE::ARRAY);
#endif //#ifdef _DEBUG

    return std::get<Array>(m_Data);
  }

#pragma warning(suppress: 26440)
  [[nodiscard]] inline const Object& AsObject() const
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_Type == TYPE::OBJECT);
#endif //#ifdef _DEBUG

    return std::get<Object>(m_Data);
  }

#pragma warning(suppress: 26440)
  [[nodiscard]] inline Object& AsObject()
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_Type == TYPE::OBJECT);
#endif //#ifdef _DEBUG

    return std::get<Object>(m_Data);
  }

#pragma warning(suppress: 26440)
  [[nodiscard]] inline bool AsBoolean() const
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_Type == TYPE::BOOLEAN);
#endif //#ifdef _DEBUG

    return std::get<bool>(m_Data);
  }

#pragma warning(suppress: 26440)
  [[nodiscard]] inline bool& AsBoolean()
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_Type == TYPE::BOOLEAN);
#endif //#ifdef _DEBUG

    return std::get<bool>(m_Data);
  }

  void Parse(_In_z_ const char* pszJSON, _In_ int nStringReserveSize = -1, _In_ std::uint32_t nFlags = 0)
  {
    std::wstring sWideString{UTF82W(pszJSON)};

    //Delegate to the other version of the method
    Parse(sWideString.c_str(), nStringReserveSize, nFlags);
  }

  void Parse(_In_z_ const wchar_t* pszJSON, _In_ int nStringReserveSize = -1, _In_ std::uint32_t nFlags = 0)
  {
    const wchar_t* pszCurrentChar{pszJSON};

    if (nFlags & PARSE_FLAGS::PARSE_JSON5)
    {
      //Skip any leading whitespace or comments
      SkipWhitespaceOrCommentNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON, nFlags, pszCurrentChar);
    }
    else
    {
      //Skip any leading whitespace
      SkipWhitespaceNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON, nFlags, pszCurrentChar);
    }

    //Delegate the main work to the helper method
    ParseHelper(pszJSON, pszCurrentChar, nStringReserveSize, nFlags);

    if (nFlags & PARSE_FLAGS::PARSE_JSON5)
    {
      //Should only be whitespace or comments left now
      SkipWhitespaceOrCommentOnlyNullOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON, nFlags, pszCurrentChar);
    }
    else
    {
      //Should only be whitespace left now
      SkipWhitespaceOnlyNullOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON, nFlags, pszCurrentChar);
    }
  }

  CValue* JSONPointer(_In_z_ const char* pszJSONPointer)
  {
    std::wstring sWideString{UTF82W(pszJSONPointer)};

    //Delegate to the other version of the method
#pragma warning(suppress: 26487)
    return JSONPointer(sWideString.c_str());
  }

  CValue* JSONPointer(_In_z_ const wchar_t* pszJSONPointer)
  {
    //What will be the return value from this method
#pragma warning(suppress: 26429)
    CValue* pCurrent{this};

    //Handle the special case of an empty string
#pragma warning(suppress: 26429)
    const wchar_t* pszCurrentChar{pszJSONPointer};
    if (*pszCurrentChar == L'\0')
      return pCurrent;

    bool bEndOfTokens{false};
    while (!bEndOfTokens)
    {
      //Skip over the "/"
      if (*pszCurrentChar != L'/')
#pragma warning(suppress: 26472)
        throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON_POINTER, static_cast<std::size_t>(pszCurrentChar - pszJSONPointer)};
#pragma warning(suppress: 26481)
      ++pszCurrentChar;

      //Extract the next JSON pointer token
#pragma warning(suppress: 26486)
      std::wstring sToken{ExtractJSONPointerString(pszJSONPointer, pszCurrentChar)};

      //Navigate to the next node based on whether the current node is a JSON object or a JSON array
      if (pCurrent->IsObject())
      {
        if (!pCurrent->HasChild(sToken.c_str()))
#pragma warning(suppress: 26472)
          throw CParseException{PARSE_EXCEPTION_REASON::COULD_NOT_FIND_CHILD_IN_OBJECT, static_cast<std::size_t>(pszCurrentChar - pszJSONPointer)};
        pCurrent = &pCurrent->Child(sToken.c_str());
      }
      else if (pCurrent->IsArray())
      {
        const int nIndex{ExtractJSONPointerInt(sToken.c_str())};
        if (!pCurrent->HasChild(nIndex))
#pragma warning(suppress: 26472)
          throw CParseException{PARSE_EXCEPTION_REASON::COULD_NOT_FIND_CHILD_IN_ARRAY, static_cast<std::size_t>(pszCurrentChar - pszJSONPointer)};
        pCurrent = &pCurrent->Child(nIndex);
      }
      else
      {
#pragma warning(suppress: 26472)
        throw CParseException{PARSE_EXCEPTION_REASON::CURRENT_ITEM_IS_NOT_OBJECT_OR_ARRAY, static_cast<std::size_t>(pszCurrentChar - pszJSONPointer)};
      }

      //Prepare for the next time around
      bEndOfTokens = (*pszCurrentChar == L'\0');
    }

    return pCurrent;
  }

  [[nodiscard]] std::wstring Encode(_In_ std::uint32_t nFlags = ENCODE_FLAGS::NONE) const
  {
    return Encode(nFlags, 0);
  }

  [[nodiscard]] std::size_t CountChildren() const
  {
    if (m_Type == TYPE::OBJECT)
      return AsObject().size();
    else if (m_Type == TYPE::ARRAY)
      return AsArray().size();
    else
      return 0;
  }

  [[nodiscard]] bool HasChild(_In_ std::size_t nIndex) const
  {
    if (m_Type == TYPE::ARRAY)
      return nIndex < AsArray().size();
    else
      return false;
  }

  [[nodiscard]] bool HasChild(_In_z_ const wchar_t* pszName) const
  {
    if (m_Type == TYPE::OBJECT)
    {
      const Object& object{AsObject()};
      return object.find(pszName) != object.cend();
    }
    else
      return false;
  }

  [[nodiscard]] CValue& Child(_In_ std::size_t nIndex)
  {
    //Validate our parameters
  #ifdef _DEBUG
    assert(m_Type == TYPE::ARRAY);
  #endif

#pragma warning(suppress: 26446)
    return AsArray()[nIndex];
  }

  [[nodiscard]] const CValue& Child(std::size_t nIndex) const
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_Type == TYPE::ARRAY);
#endif

#pragma warning(suppress: 26446)
    return AsArray()[nIndex];
  }

  [[nodiscard]] CValue& Child(_In_z_ const wchar_t* pszName)
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_Type == TYPE::OBJECT);
#endif

#pragma warning(suppress: 26487 26489)
    return AsObject().find(pszName)->second;
  }

  [[nodiscard]] const CValue& Child(_In_z_ const wchar_t* pszName) const
  {
    //Validate our parameters
#ifdef _DEBUG
    assert(m_Type == TYPE::OBJECT);
#endif

#pragma warning(suppress: 26487 26489)
    return AsObject().find(pszName)->second;
  }

  [[nodiscard]] std::vector<std::wstring> ObjectKeys() const
  {
    //What will be the return value from this method
    std::vector<std::wstring> keys;

    if (m_Type == TYPE::OBJECT)
    {
      const Object& object{AsObject()};
      keys.reserve(object.size());
      for (auto& iter : object)
#pragma warning(suppress: 26489)
        keys.push_back(iter.first);
    }

    return keys;
  }

  static std::wstring UTF82W(_In_z_ const char* pszText)
  {
    //First call the function to determine how much space we need to allocate
    int nWideLength{MultiByteToWideChar(CP_UTF8, 0, pszText, -1, nullptr, 0)};

    //Now recall with the buffer to get the converted text
    std::wstring sWideString;
    sWideString.resize(nWideLength);
    nWideLength = MultiByteToWideChar(CP_UTF8, 0, pszText, -1, sWideString.data(), nWideLength);
    sWideString.resize(nWideLength);

    return sWideString;
  }

protected:
//Member variables
  std::variant<bool, double, std::wstring, Object, Array> m_Data;
  TYPE m_Type;

//Methods
#pragma warning(suppress: 26429)
  void ParseHelper(_In_z_ const wchar_t* pszJSON, _Inout_ const wchar_t*& pszCurrentChar, _In_ int nStringReserveSize, _In_ std::uint32_t nFlags)
  {
    const size_t nLen{wcslen(pszCurrentChar)};
    if (*pszCurrentChar == L'"') //Is the data a JSON double quoted string
    {
#pragma warning(suppress: 26481 26486)
      *this = ExtractString(pszJSON, L'"', nFlags, ++pszCurrentChar, nStringReserveSize);
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (*pszCurrentChar == L'\'')) //Is the data a JSON5 single quoted string and we are parsing JSON5
    {
#pragma warning(suppress: 26481 26486)
      *this = ExtractString(pszJSON, L'\'', nFlags, ++pszCurrentChar, nStringReserveSize);
      return;
    }
    else if ((nLen >= 4) && (_wcsnicmp(pszCurrentChar, L"true", 4) == 0)) //Is the data true
    {
      *this = true;
#pragma warning(suppress: 26481 26487)
      pszCurrentChar += 4;
      return;
    }
    else if ((nLen >= 5) && (_wcsnicmp(pszCurrentChar, L"false", 5) == 0)) //Is the data false
    {
      *this = false;
#pragma warning(suppress: 26481)
      pszCurrentChar += 5;
      return;
    }
    else if ((nLen >= 4) && (_wcsnicmp(pszCurrentChar, L"null", 4) == 0)) //Is the data null
    {
      SetNull();
#pragma warning(suppress: 26481)
      pszCurrentChar += 4;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 8) && (_wcsnicmp(pszCurrentChar, L"infinity", 8) == 0)) //Is the data Infinity
    {
      *this = std::numeric_limits<double>::infinity();
#pragma warning(suppress: 26481)
      pszCurrentChar += 8;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 9) && (_wcsnicmp(pszCurrentChar, L"+infinity", 9) == 0)) //Is the data Positive Infinity
    {
      *this = std::numeric_limits<double>::infinity();
#pragma warning(suppress: 26481)
      pszCurrentChar += 9;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 3) && (_wcsnicmp(pszCurrentChar, L"inf", 3) == 0)) //Is the data Infinity
    {
      *this = std::numeric_limits<double>::infinity();
#pragma warning(suppress: 26481)
      pszCurrentChar += 3;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 4) && (_wcsnicmp(pszCurrentChar, L"+inf", 4) == 0)) //Is the data Positive Infinity
    {
      *this = std::numeric_limits<double>::infinity();
#pragma warning(suppress: 26481)
      pszCurrentChar += 4;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 9) && (_wcsnicmp(pszCurrentChar, L"-infinity", 9) == 0)) //Is the data Negative Infinity
    {
      *this = -std::numeric_limits<double>::infinity();
#pragma warning(suppress: 26481)
      pszCurrentChar += 9;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 4) && (_wcsnicmp(pszCurrentChar, L"-inf", 4) == 0)) //Is the data Negative Infinity
    {
      *this = -std::numeric_limits<double>::infinity();
#pragma warning(suppress: 26481)
      pszCurrentChar += 4;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 3) && (_wcsnicmp(pszCurrentChar, L"nan", 3) == 0)) //Is the data NaN
    {
      *this = std::numeric_limits<double>::quiet_NaN();
#pragma warning(suppress: 26481)
      pszCurrentChar += 3;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 4) && (_wcsnicmp(pszCurrentChar, L"qnan", 4) == 0)) //Is the data NaN
    {
      *this = std::numeric_limits<double>::quiet_NaN();
#pragma warning(suppress: 26481)
      pszCurrentChar += 4;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 4) && (_wcsnicmp(pszCurrentChar, L"snan", 4) == 0)) //Is the data NaN
    {
      *this = std::numeric_limits<double>::quiet_NaN();
#pragma warning(suppress: 26481)
      pszCurrentChar += 4;
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 4) && (_wcsnicmp(pszCurrentChar, L"-0x", 3) == 0)) //Is the data a negative JSON5 hexadecimal number
    {
      *this = ExtractHexNumber(pszJSON, pszCurrentChar);
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 4) && (_wcsnicmp(pszCurrentChar, L"+0x", 3) == 0)) //Is the data a positive JSON5 hexadecimal number
    {
      *this = ExtractHexNumber(pszJSON, pszCurrentChar);
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (nLen >= 3) && (_wcsnicmp(pszCurrentChar, L"0x", 2) == 0)) //Is the data a JSON5 hexadecimal number
    {
      *this = ExtractHexNumber(pszJSON, pszCurrentChar);
      return;
    }
    else if ((*pszCurrentChar == L'-') || (*pszCurrentChar >= L'0' && *pszCurrentChar <= L'9')) //Is the data a number
    {
      *this = ExtractNumber(pszJSON, nFlags, pszCurrentChar);
      return;
    }
    else if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && ((*pszCurrentChar == L'.') || (*pszCurrentChar == L'+'))) //Is the data a JSON5 number with a leading '.' or '+'
    {
      *this = ExtractNumber(pszJSON, nFlags, pszCurrentChar);
      return;
    }
    else if (*pszCurrentChar == L'{') //Is the data an object
    {
#pragma warning(suppress: 26481 26487)
      ++pszCurrentChar;

      Object object;
#pragma warning(suppress: 26489)
      while (*pszCurrentChar != L'\0')
      {
        if (nFlags & PARSE_FLAGS::PARSE_JSON5)
        {
          //Skip whitespace or comment
          SkipWhitespaceOrCommentNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, nFlags, pszCurrentChar);
        }
        else
        {
          //Skip whitespace
#pragma warning(suppress: 26486)
          SkipWhitespaceNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, nFlags, pszCurrentChar);
        }

        //Handle the special case of an empty object
        if ((object.size() == 0) && (*pszCurrentChar == L'}'))
        {
#pragma warning(suppress: 26481)
          ++pszCurrentChar;
          *this = std::move(object);
          return;
        }

        //Parse the child element name now
        bool bParsingString = true;
        wchar_t cEndDelimiter = L'\0';
        if (nFlags & PARSE_FLAGS::PARSE_JSON5)
        {
          bParsingString = (*pszCurrentChar == L'\"') || (*pszCurrentChar == L'\'');
          if (bParsingString)
          {
            cEndDelimiter = *pszCurrentChar;
#pragma warning(suppress: 26481)
            ++pszCurrentChar;
          }
        }
        else
        {
          if (*pszCurrentChar != L'\"')
#pragma warning(suppress: 26472)
            throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
#pragma warning(suppress: 26481)
          ++pszCurrentChar;
          cEndDelimiter = L'\"';
        }

        std::wstring sName;
        if (bParsingString)
          sName = ExtractString(pszJSON, cEndDelimiter, nFlags, pszCurrentChar, nStringReserveSize);
        else
          sName = ExtractJSON5Identifier(pszJSON, nFlags, pszCurrentChar, nStringReserveSize);

        if (nFlags & PARSE_FLAGS::PARSE_JSON5)
        {
          //Skip whitespace or comment
          SkipWhitespaceOrCommentNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, nFlags, pszCurrentChar);
        }
        else
        {
          //Skip whitespace
          SkipWhitespaceNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, nFlags, pszCurrentChar);
        }

        //There should be a ":" now
        if (*pszCurrentChar != L':')
#pragma warning(suppress: 26472)
          throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
#pragma warning(suppress: 26481)
        ++pszCurrentChar;

        if (nFlags & PARSE_FLAGS::PARSE_JSON5)
        {
          //Skip whitespace or comment
#pragma warning(suppress: 26486)
          SkipWhitespaceOrCommentNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, nFlags, pszCurrentChar);
        }
        else
        {
          //Skip whitespace
#pragma warning(suppress: 26486)
          SkipWhitespaceNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, nFlags, pszCurrentChar);
        }

        //Parse the child element value now
        CValue childElement;
        childElement.ParseHelper(pszJSON, pszCurrentChar, nStringReserveSize, nFlags);
        object[std::move(sName)] = std::move(childElement);

        if (nFlags & PARSE_FLAGS::PARSE_JSON5)
        {
          //Skip whitespace or comment
          SkipWhitespaceOrCommentNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, nFlags, pszCurrentChar);
        }
        else
        {
          //Skip whitespace
          SkipWhitespaceNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, nFlags, pszCurrentChar);
        }

        //Are we at the end of the object yet?
        if (*pszCurrentChar == L'}')
        {
#pragma warning(suppress: 26481 26487)
          ++pszCurrentChar;
          *this = std::move(object);
          return;
        }

        //There should be a comma now
        if (*pszCurrentChar != L',')
#pragma warning(suppress: 26472)
          throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

        if (nFlags & PARSE_FLAGS::PARSE_JSON5)
        {
          //Check to see if we have a trailing "," in the object
#pragma warning(suppress: 26429)
          auto pszTempCurrentChar = pszCurrentChar;
#pragma warning(suppress: 26481)
          ++pszTempCurrentChar;
          SkipWhitespaceOrCommentNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_OBJECT, nFlags, pszTempCurrentChar);
          if (*pszTempCurrentChar == L'}')
          {
#pragma warning(suppress: 26481)
            pszCurrentChar = ++pszTempCurrentChar;
            *this = std::move(object);
            return;
          }
        }

        //Prepare for the next time around
#pragma warning(suppress: 26481)
        ++pszCurrentChar;
      }
    }
    else if (*pszCurrentChar == L'[') //Is the data an array
    {
#pragma warning(suppress: 26481 26487)
      ++pszCurrentChar;

      Array array;
#pragma warning(suppress: 26489)
      while (*pszCurrentChar != L'\0')
      {
        if (nFlags & PARSE_FLAGS::PARSE_JSON5)
        {
          //Skip any leading whitespace or comments
#pragma warning(suppress: 26486)
          SkipWhitespaceOrCommentNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ARRAY, nFlags, pszCurrentChar);
        }
        else
        {
          //Skip any leading whitespace
#pragma warning(suppress: 26486)
          SkipWhitespaceNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ARRAY, nFlags, pszCurrentChar);
        }

        //Handle the special case of an empty array
        if ((array.size() == 0) && (*pszCurrentChar == L']'))
        {
#pragma warning(suppress: 26481)
          ++pszCurrentChar;
          *this = std::move(array);
          return;
        }

        //Parse the child element value now
        CValue childElement;
        childElement.ParseHelper(pszJSON, pszCurrentChar, nStringReserveSize, nFlags);
        array.push_back(std::move(childElement));

        if (nFlags & PARSE_FLAGS::PARSE_JSON5)
        {
          //Skip any trailing whitespace or comments
          SkipWhitespaceOrCommentNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ARRAY, nFlags, pszCurrentChar);
        }
        else
        {
          //Skip any trailing whitespace
          SkipWhitespaceNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ARRAY, nFlags, pszCurrentChar);
        }

        //Are we at the end of the array yet?
        if (*pszCurrentChar == L']')
        {
#pragma warning(suppress: 26481 26487)
          ++pszCurrentChar;
          *this = std::move(array);
          return;
        }

        //There should be a comma now
        if (*pszCurrentChar != L',')
#pragma warning(suppress: 26472)
          throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ARRAY, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

        if (nFlags & PARSE_FLAGS::PARSE_JSON5)
        {
          //Check to see if we have a trailing "," in the array
#pragma warning(suppress: 26429)
          auto pszTempCurrentChar = pszCurrentChar;
#pragma warning(suppress: 26481)
          ++pszTempCurrentChar;
          SkipWhitespaceOrCommentNullNotOK(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ARRAY, nFlags, pszTempCurrentChar);
          if (*pszTempCurrentChar == L']')
          {
#pragma warning(suppress: 26481)
            pszCurrentChar = ++pszTempCurrentChar;
            *this = std::move(array);
            return;
          }
        }

        //Prepare for the next time around
#pragma warning(suppress: 26481)
        ++pszCurrentChar;
      }
    }
    else
#pragma warning(suppress: 26472)
      throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
  }

  static void HandleBeforeValue(_Inout_ std::wstring& s, _In_ std::uint32_t nFlags)
  {
    if (nFlags & WHITESPACE_BEFORE_VALUE)
    {
      if (nFlags & TABS)
        s += L"\t";
      else if (nFlags & FOUR_SPACES)
        s += L"    ";
      else if (nFlags & THREE_SPACES)
        s += L"   ";
      else if (nFlags & TWO_SPACES)
        s += L"  ";
      else if (nFlags & ONE_SPACE)
        s += L" ";
    }
  }

  static void HandleAfterValue(_Inout_ std::wstring& s, _In_ std::uint32_t nFlags)
  {
    if (nFlags & WHITESPACE_AFTER_VALUE)
    {
      if (nFlags & TABS)
        s += L"\t";
      else if (nFlags & FOUR_SPACES)
        s += L"    ";
      else if (nFlags & THREE_SPACES)
        s += L"   ";
      else if (nFlags & TWO_SPACES)
        s += L"  ";
      else if (nFlags & ONE_SPACE)
        s += L" ";
    }
  }

  static void EncodeEncodedDigit(_Inout_ std::wstring& s, _In_ int nDigit)
  {
    if ((nDigit >= 0) && (nDigit <= 9))
#pragma warning(suppress: 26472)
      s += static_cast<wchar_t>(L'0' + nDigit);
    else
#pragma warning(suppress: 26472)
      s += static_cast<wchar_t>(L'A' + (nDigit - 10));
  }

  static void HandleEncodeCharacter(_Inout_ std::wstring& s, _In_ wchar_t chr)
  {
    s += L"\\u";
    EncodeEncodedDigit(s, (chr & 0xF000) >> 12);
#pragma warning(suppress: 26486 26489)
    EncodeEncodedDigit(s, (chr & 0x0F00) >> 8);
#pragma warning(suppress: 26486 26489)
    EncodeEncodedDigit(s, (chr & 0x00F0) >> 4);
#pragma warning(suppress: 26486 26489)
    EncodeEncodedDigit(s, chr & 0x000F);
  }

  static std::wstring EncodeString(_In_ const std::wstring& string, _In_ std::uint32_t nFlags)
  {
    std::wstring sReturn;
    sReturn.reserve(string.size() + 2);
    sReturn = L"";
    HandleBeforeValue(sReturn, nFlags);
    sReturn += L"\"";
    for (auto& chr : string)
    {
      if (nFlags & ENCODE_ALL_UNICODE_CHARACTERS)
        HandleEncodeCharacter(sReturn, chr);
      else
      {
        if (chr == L'\"')
          sReturn += L"\\\"";
        else if (chr == L'\\')
          sReturn += L"\\\\";
        else if (chr == L'/')
          sReturn +=  L"\\/";
        else if (chr == L'\b')
          sReturn += L"\\b";
        else if (chr == L'\f')
          sReturn += L"\\f";
        else if (chr == L'\n')
          sReturn += L"\\n";
        else if (chr == L'\r')
          sReturn += L"\\r";
        else if (chr == L'\t')
          sReturn += L"\\t";
        else if (((chr < L' ') || (chr > L'~')) && (nFlags & ENCODE_NON_ASCII_CHARACTERS))
          HandleEncodeCharacter(sReturn, chr);
        else
          sReturn += chr;
      }
    }
    sReturn += L"\"";
    HandleAfterValue(sReturn, nFlags);
    return sReturn;
  }

  static constexpr bool CurrentCharIsWhitespace(_In_ wchar_t cCurrentChar, _In_ std::uint32_t nFlags) noexcept
  {
    if (nFlags & PARSE_FLAGS::PARSE_JSON5)
    {
      return ((cCurrentChar == L'\x0009') || (cCurrentChar == L'\x000A') || (cCurrentChar == L'\x000B') || (cCurrentChar == L'\x000C') ||
              (cCurrentChar == L'\x000D') || (cCurrentChar == L'\x0020') || (cCurrentChar == L'\x00A0') || (cCurrentChar == L'\x2028') ||
              (cCurrentChar == L'\x2029') || (cCurrentChar == L'\xFEFF') || (cCurrentChar == L'\x1680') ||
              ((cCurrentChar >= L'\x2000') && (cCurrentChar <= L'\x200A')) || (cCurrentChar == L'\x202F') ||
              (cCurrentChar == L'\x205F') || (cCurrentChar == L'\x3000'));
    }
    else
      return (cCurrentChar == L' ' || cCurrentChar == L'\t' || cCurrentChar == L'\r' || cCurrentChar == L'\n');
  }

#pragma warning(suppress: 26429)
  static constexpr bool CurrentCharIsStartSingleLineComment(_In_z_ const wchar_t* pszJSON) noexcept
  {
    if (*pszJSON != L'/')
      return false;
#pragma warning(suppress: 26481)
    ++pszJSON;
    return *pszJSON == L'/';
  }

#pragma warning(suppress: 26429)
  static constexpr bool CurrentCharIsStartMultiLineComment(_In_z_ const wchar_t* pszJSON) noexcept
  {
    if (*pszJSON != L'/')
      return false;
#pragma warning(suppress: 26481)
    ++pszJSON;
    return *pszJSON == L'*';
  }

#pragma warning(suppress: 26429)
  static constexpr bool CurrentCharIsEOL(_In_ wchar_t cCurrentChar) noexcept
  {
    return (cCurrentChar == L'\x2028' || cCurrentChar == L'\x2029' || cCurrentChar == L'\r' || cCurrentChar == L'\n');
  }

#pragma warning(suppress: 26429)
  static constexpr bool CurrentCharIsHexDigit(_In_ wchar_t cCurrentChar) noexcept
  {
    return ((cCurrentChar >= L'0') && (cCurrentChar <= L'9')) || ((cCurrentChar >= L'A') && (cCurrentChar <= L'F')) || ((cCurrentChar >= L'a') && (cCurrentChar <= L'f'));
  }

#pragma warning(suppress: 26429)
  static void SkipWhitespace(_In_ std::uint32_t nFlags, _Inout_ const wchar_t*& pszCurrentChar) noexcept
  {
#pragma warning(suppress: 26489)
    while ((*pszCurrentChar != L'\0') && CurrentCharIsWhitespace(*pszCurrentChar, nFlags))
#pragma warning(suppress: 26481 26487)
      ++pszCurrentChar;
  }

#pragma warning(suppress: 26429)
  static void SkipWhitespaceNullNotOK(_In_z_ const wchar_t* pszJSON, _In_ PARSE_EXCEPTION_REASON parseError, _In_ std::uint32_t nFlags, _Inout_ const wchar_t*& pszCurrentChar)
  {
    SkipWhitespace(nFlags, pszCurrentChar);

#pragma warning(suppress: 26489)
    if (*pszCurrentChar == L'\0')
#pragma warning(suppress: 26472)
      throw CParseException{parseError, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
  }

#pragma warning(suppress: 26429)
  static void SkipWhitespaceOnlyNullOK(_In_z_ const wchar_t* pszJSON, _In_ PARSE_EXCEPTION_REASON parseError, _In_ std::uint32_t nFlags, _Inout_ const wchar_t*& pszCurrentChar)
  {
    SkipWhitespace(nFlags, pszCurrentChar);

#pragma warning(suppress: 26489)
    if (*pszCurrentChar != L'\0')
#pragma warning(suppress: 26472)
      throw CParseException{parseError, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
  }

#pragma warning(suppress: 26429 26440)
  static void SkipSingleLineComment(_In_z_ const wchar_t* /*pszJSON*/, _Inout_ const wchar_t*& pszCurrentChar)
  {
#pragma warning(suppress: 26489)
    while ((*pszCurrentChar != L'\0') && !CurrentCharIsEOL(*pszCurrentChar))
#pragma warning(suppress: 26481 26487)
      ++pszCurrentChar;
#pragma warning(suppress: 26481)
    if ((*pszCurrentChar == L'\r') && (*(pszCurrentChar + 1) == L'\n'))
    {
#pragma warning(suppress: 26481 26487)
        ++pszCurrentChar;
    }
#pragma warning(suppress: 26481 26487)
    ++pszCurrentChar;
  }

#pragma warning(suppress: 26429)
  static void SkipMultiLineComment(_In_z_ const wchar_t* pszJSON, _Inout_ const wchar_t*& pszCurrentChar)
  {
    bool bFoundEndComment = false;
    bool bPrevCharWasAsterisk = false;
    bool bCurrentCharIsAsterisk = false;
    bool bCurrentCharIsSolidus = false;
    while (!bFoundEndComment)
    {
      bPrevCharWasAsterisk = bCurrentCharIsAsterisk;
      if (*pszCurrentChar == L'*')
        bCurrentCharIsAsterisk = true;
      else if (*pszCurrentChar == L'/')
        bCurrentCharIsSolidus = true;
#pragma warning(suppress: 26489)
      else if (*pszCurrentChar == L'\0')
#pragma warning(suppress: 26472)
        throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_MULTI_LINE_COMMENT, static_cast<std::size_t>(pszCurrentChar - pszJSON) };
      if (bPrevCharWasAsterisk && bCurrentCharIsSolidus)
        bFoundEndComment = true;
#pragma warning(suppress: 26481 26487)
      ++pszCurrentChar;
    }
  }

#pragma warning(suppress: 26429)
  static void SkipWhitespaceOrComment(_In_z_ const wchar_t* pszJSON, _In_ std::uint32_t nFlags, _Inout_ const wchar_t*& pszCurrentChar)
  {
    bool bContinue{true};
    while (bContinue)
    {
      if (CurrentCharIsWhitespace(*pszCurrentChar, nFlags))
        SkipWhitespace(nFlags, pszCurrentChar);
      else if (CurrentCharIsStartMultiLineComment(pszCurrentChar))
      {
#pragma warning(suppress: 26481)
        pszCurrentChar += 2;
        SkipMultiLineComment(pszJSON, pszCurrentChar);
      }
      else if (CurrentCharIsStartSingleLineComment(pszCurrentChar))
      {
#pragma warning(suppress: 26481)
        pszCurrentChar += 2;
        SkipSingleLineComment(pszJSON, pszCurrentChar);
      }
      else
        bContinue = false;
    }
  }

#pragma warning(suppress: 26429)
  static void SkipWhitespaceOrCommentNullNotOK(_In_z_ const wchar_t* pszJSON, _In_ PARSE_EXCEPTION_REASON parseError, _In_ std::uint32_t nFlags, _Inout_ const wchar_t*& pszCurrentChar)
  {
    SkipWhitespaceOrComment(pszJSON, nFlags, pszCurrentChar);

#pragma warning(suppress: 26489)
    if (*pszCurrentChar == L'\0')
#pragma warning(suppress: 26472)
      throw CParseException{parseError, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
  }

#pragma warning(suppress: 26429)
  static void SkipWhitespaceOrCommentOnlyNullOK(_In_z_ const wchar_t* pszJSON, _In_ PARSE_EXCEPTION_REASON parseError, _In_ std::uint32_t nFlags, _Inout_ const wchar_t*& pszCurrentChar)
  {
    SkipWhitespaceOrComment(pszJSON, nFlags, pszCurrentChar);

#pragma warning(suppress: 26489)
    if (*pszCurrentChar != L'\0')
#pragma warning(suppress: 26472)
      throw CParseException{parseError, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
  }

#pragma warning(suppress: 26429)
  static double ExtractHexInt(_In_z_ const wchar_t* pszJSON, _Inout_ const wchar_t*& pszCurrentChar)
  {
    //We need at least one hex digit for an hex int
    if (!CurrentCharIsHexDigit(*pszCurrentChar))
#pragma warning(suppress: 26472)
      throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_NUMBER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

    //What will be the return value from this function
    double fNumber{0};

#pragma warning(suppress: 26489)
    while (CurrentCharIsHexDigit(*pszCurrentChar))
    {
      if ((*pszCurrentChar >= L'0') && (*pszCurrentChar <= L'9'))
        fNumber = (fNumber * 16) + (*pszCurrentChar - L'0');
      else if ((*pszCurrentChar >= L'A') && (*pszCurrentChar <= L'F'))
        fNumber = (fNumber * 16) + (*pszCurrentChar - L'A' + 10.0);
      else
        fNumber = (fNumber * 16) + (*pszCurrentChar - L'a' + 10.0);

      //Prepare for the next time around
#pragma warning(suppress: 26481 26487)
      ++pszCurrentChar;
    }

    return fNumber;
  }

#pragma warning(suppress: 26429)
  static double ExtractInt(_In_z_ const wchar_t* pszJSON, _Inout_ const wchar_t*& pszCurrentChar)
  {
    //We need at least one digit for an int
    if ((*pszCurrentChar < L'0') || (*pszCurrentChar > L'9'))
#pragma warning(suppress: 26472)
      throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_NUMBER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

    //What will be the return value from this function
    double fNumber{0};

#pragma warning(suppress: 26489)
    while ((*pszCurrentChar >= L'0') && (*pszCurrentChar <= L'9'))
    {
#pragma warning(suppress: 26489)
      fNumber = (fNumber * 10) + (*pszCurrentChar - L'0');

      //Prepare for the next time around
#pragma warning(suppress: 26481 26487)
      ++pszCurrentChar;
    }

    return fNumber;
  }

#pragma warning(suppress: 26429)
  static double ExtractDecimal(_In_z_ const wchar_t* pszJSON, _In_ PARSE_EXCEPTION_REASON parseError, _Inout_ const wchar_t*& pszCurrentChar)
  {
    //We need at least one digit for a decimal
    if ((*pszCurrentChar < L'0') || (*pszCurrentChar > L'9'))
#pragma warning(suppress: 26472)
      throw CParseException{parseError, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

    //What will be the return value from this function
    double fDecimal{0};

    double fFactor{0.1};
#pragma warning(suppress: 26489)
    while ((*pszCurrentChar >= L'0') && (*pszCurrentChar <= L'9'))
    {
#pragma warning(suppress: 26489)
      const int nDigit{*pszCurrentChar - L'0'};
      fDecimal = fDecimal + nDigit*fFactor;

      //Prepare for the next time around
#pragma warning(suppress: 26481 26487)
      ++pszCurrentChar;
      fFactor *= 0.1;
    }

    return fDecimal;
  }

#pragma warning(suppress: 26429)
  static double ExtractHexNumber(_In_z_ const wchar_t* pszJSON, _Inout_ const wchar_t*& pszCurrentChar)
  {
    //Handle the optional plus or minus sign
    bool bNegative{false};
    if (*pszCurrentChar == L'-')
    {
      bNegative = true;
#pragma warning(suppress: 26481)
      ++pszCurrentChar;
    }
    else if (*pszCurrentChar == L'+')
    {
#pragma warning(suppress: 26481)
      ++pszCurrentChar;
    }

    //Skip over the "0x" prefix
    if (*pszCurrentChar != L'0')
#pragma warning(suppress: 26472)
      throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_NUMBER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
#pragma warning(suppress: 26481)
    ++pszCurrentChar;
    if ((*pszCurrentChar != L'x') && (*pszCurrentChar != L'X'))
#pragma warning(suppress: 26472)
      throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_NUMBER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
#pragma warning(suppress: 26481)
    ++pszCurrentChar;

    double fNumber{ExtractHexInt(pszJSON, pszCurrentChar)};

    //Handle the case of the number being negative
    if (bNegative)
      fNumber *= -1;

    return fNumber;
  }

#pragma warning(suppress: 26429)
  static double ExtractNumber(_In_z_ const wchar_t* pszJSON, _In_ std::uint32_t nFlags, _Inout_ const wchar_t*& pszCurrentChar)
  {
    //Handle the optional plus or minus sign
    bool bNegative{false};
    if (*pszCurrentChar == L'-')
    {
      bNegative = true;
#pragma warning(suppress: 26481)
      ++pszCurrentChar;
    }
    else if (*pszCurrentChar == L'+')
    {
      //Leading + is only valid for JSON5
      if (nFlags & PARSE_FLAGS::PARSE_JSON5)
      {
#pragma warning(suppress: 26481)
        ++pszCurrentChar;
      }
      else
#pragma warning(suppress: 26472)
        throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_NUMBER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
    }

    double fNumber{0};
    if ((nFlags & PARSE_FLAGS::PARSE_JSON5) && (*pszCurrentChar == L'.'))
    {
      //The whole part of the number is optional for JSON5 numbers
    }
    else
    {
      //Parse the whole part of the number
  #pragma warning(suppress: 26489)
      if (*pszCurrentChar == L'0')
  #pragma warning(suppress: 26481 26487)
        ++pszCurrentChar;
      else
  #pragma warning(suppress: 26486)
        fNumber = ExtractInt(pszJSON, pszCurrentChar);
    }

    //There could be a decimal point now
#pragma warning(suppress: 26489)
    if (*pszCurrentChar == L'.')
    {
#pragma warning(suppress: 26481)
      ++pszCurrentChar;

      bool bDecimalPart{true};
      if (nFlags & PARSE_FLAGS::PARSE_JSON5)
      {
        //The decimal part of the number is optional for JSON5 numbers
        if ((*pszCurrentChar < L'0') || (*pszCurrentChar > L'9'))
          bDecimalPart = false;
      }

      if (bDecimalPart)
      {
        //Pull out the decimal part and add it to the number
#pragma warning(suppress: 26486)
        fNumber += ExtractDecimal(pszJSON, PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_NUMBER, pszCurrentChar);
      }
    }

    //There could be an exponent now
#pragma warning(suppress: 26489)
    if (*pszCurrentChar == L'e' || *pszCurrentChar == L'E')
    {
#pragma warning(suppress: 26481)
      ++pszCurrentChar;

      //Check the sign of the exponent
      bool bNegativeExponent{ false };
#pragma warning(suppress: 26489)
      if ((*pszCurrentChar == L'-') || (*pszCurrentChar == L'+'))
      {
#pragma warning(suppress: 26489)
        bNegativeExponent = (*pszCurrentChar == L'-');
#pragma warning(suppress: 26481)
        ++pszCurrentChar;
      }

      //Parse the exponent value
#pragma warning(suppress: 26486)
      const int nExponent{static_cast<int>(ExtractInt(pszJSON, pszCurrentChar))};
      for (int i=0; i<nExponent; i++)
        fNumber = bNegativeExponent ? (fNumber / 10.0) : (fNumber * 10.0);
    }

    //Handle the case of the number being negative
    if (bNegative)
      fNumber *= -1;

    return fNumber;
  }

#pragma warning(suppress: 26429)
  static std::wstring ExtractString(_In_z_ const wchar_t* pszJSON, _In_ wchar_t cEndDelimiter, _In_ std::uint32_t nFlags, _Inout_ const wchar_t*& pszCurrentChar, _In_ int nStringReserveSize)
  {
    //What will be the return value from this method
    std::wstring sReturn;
    if (nStringReserveSize != -1)
      sReturn.reserve(nStringReserveSize);
    sReturn = L"";

#pragma warning(suppress: 26489)
    while (*pszCurrentChar != L'\0')
    {
#pragma warning(suppress: 26489)
      wchar_t currentChar{*pszCurrentChar};

      if (currentChar == L'\\') //The string has been escaped
      {
#pragma warning(suppress: 26481)
        ++pszCurrentChar;

        switch (*pszCurrentChar)
        {
          case L'"':
          {
            currentChar = L'"';
            break;
          }
          case L'\\':
          {
            currentChar = L'\\';
            break;
          }
          case L'/':
          {
            currentChar = L'/';
            break;
          }
          case L'b':
          {
            currentChar = L'\b';
            break;
          }
          case L'f':
          {
            currentChar = L'\f';
            break;
          }
          case L'n':
          {
            currentChar = L'\n';
            break;
          }
          case L'r':
          {
            currentChar = L'\r';
            break;
          }
          case L't':
          {
            currentChar = L'\t';
            break;
          }
          case L'u':
          {
            //We need 4 digits after the "u"
#pragma warning(suppress: 26486)
            if (wcslen(pszCurrentChar) < 5)
#pragma warning(suppress: 26472)
              throw CParseException{PARSE_EXCEPTION_REASON::ENCODED_UNICODE_CHARACTER_TOO_SHORT, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

            //Parse out the 4 digits after the "u"
            currentChar = L'\0';
            for (int i=0; i<4; i++)
            {
#pragma warning(suppress: 26481)
              ++pszCurrentChar;
              currentChar <<= 4;
#pragma warning(suppress: 26489)
              if ((*pszCurrentChar >= L'0') && (*pszCurrentChar <= L'9'))
                currentChar |= (*pszCurrentChar - L'0');
#pragma warning(suppress: 26489)
              else if ((*pszCurrentChar >= L'A') && (*pszCurrentChar <= L'F'))
                currentChar |= (10 + (*pszCurrentChar - L'A'));
#pragma warning(suppress: 26489)
              else if ((*pszCurrentChar >= L'a') && (*pszCurrentChar <= L'f'))
                currentChar |= (10 + (*pszCurrentChar - L'a'));
              else
#pragma warning(suppress: 26472)
                throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ENCODED_UNICODE_CHARACTER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
            }
            break;
          }
          default:
          {
            if (nFlags & PARSE_FLAGS::PARSE_JSON5)
            {
              //Skip over escaped line terminators
              if (CurrentCharIsEOL(*pszCurrentChar))
              {
#pragma warning(suppress: 26481)
                if ((*pszCurrentChar == L'\r') && (*(pszCurrentChar + 1) == L'\n'))
#pragma warning(suppress: 26481 26487)
                  ++pszCurrentChar;
#pragma warning(suppress: 26481)
                ++pszCurrentChar;
                continue;
              }
              else
              {
                switch (*pszCurrentChar)
                {
                  case L'v':
                  {
                    currentChar = L'\v';
                    break;
                  }
                  case L'0':
                  {
                    currentChar = L'\0';
                    break;
                  }
                  case L'x':
                  {
                    //We need 2 digits after the "x"
#pragma warning(suppress: 26486)
                    if (wcslen(pszCurrentChar) < 3)
#pragma warning(suppress: 26472)
                      throw CParseException{PARSE_EXCEPTION_REASON::ENCODED_UNICODE_CHARACTER_TOO_SHORT, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

                    //Parse out the 2 digits after the "x"
                    currentChar = L'\0';
                    for (int i=0; i<2; i++)
                    {
#pragma warning(suppress: 26481)
                      ++pszCurrentChar;
                      currentChar <<= 4;
#pragma warning(suppress: 26489)
                      if ((*pszCurrentChar >= L'0') && (*pszCurrentChar <= L'9'))
                        currentChar |= (*pszCurrentChar - L'0');
#pragma warning(suppress: 26489)
                      else if ((*pszCurrentChar >= L'A') && (*pszCurrentChar <= L'F'))
                        currentChar |= (10 + (*pszCurrentChar - L'A'));
#pragma warning(suppress: 26489)
                      else if ((*pszCurrentChar >= L'a') && (*pszCurrentChar <= L'f'))
                        currentChar |= (10 + (*pszCurrentChar - L'a'));
                      else
#pragma warning(suppress: 26472)
                        throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ENCODED_UNICODE_CHARACTER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
                    }
                    break;
                  }
                  default:
                  {
                    //decimal digits 1 through 9 are not allowed to be escaped in JSON5
                    if ((*pszCurrentChar >= L'1') && (*pszCurrentChar <= L'9'))
#pragma warning(suppress: 26472)
                      throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_STRING, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
                    currentChar = *pszCurrentChar;
                    break;
                  }
                }
              }
            }
            else
#pragma warning(suppress: 26472)
              throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_STRING, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
            break;
          }
        }
      }
      else if (currentChar == cEndDelimiter) //The end of the string
      {
#pragma warning(suppress: 26481 26487)
        ++pszCurrentChar;
        return sReturn;
      }
      else if (currentChar < L' ') //A character which is not allowed
#pragma warning(suppress: 26472)
        throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_STRING, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

      sReturn += currentChar;
#pragma warning(suppress: 26481)
      ++pszCurrentChar;
    }

#pragma warning(suppress: 26472)
    throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_STRING, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
  }

#pragma warning(suppress: 26429)
  static std::wstring ExtractJSON5Identifier(_In_z_ const wchar_t* pszJSON, _In_ std::uint32_t nFlags, _Inout_ const wchar_t*& pszCurrentChar, _In_ int nStringReserveSize)
  {
    //What will be the return value from this method
    std::wstring sReturn;
    if (nStringReserveSize != -1)
      sReturn.reserve(nStringReserveSize);
    sReturn = L"";

#pragma warning(suppress: 26489)
    while (*pszCurrentChar != L'\0')
    {
#pragma warning(suppress: 26489)
      wchar_t currentChar{*pszCurrentChar};

      if (currentChar == L'\\') //The string has been escaped
      {
#pragma warning(suppress: 26481)
        ++pszCurrentChar;

        switch (*pszCurrentChar)
        {
          case L'"':
          {
            currentChar = L'"';
            break;
          }
          case L'\\':
          {
            currentChar = L'\\';
            break;
          }
          case L'/':
          {
            currentChar = L'/';
            break;
          }
          case L'b':
          {
            currentChar = L'\b';
            break;
          }
          case L'f':
          {
            currentChar = L'\f';
            break;
          }
          case L'n':
          {
            currentChar = L'\n';
            break;
          }
          case L'r':
          {
            currentChar = L'\r';
            break;
          }
          case L't':
          {
            currentChar = L'\t';
            break;
          }
          case L'u':
          {
            //We need 4 digits after the "u"
#pragma warning(suppress: 26486)
            if (wcslen(pszCurrentChar) < 5)
#pragma warning(suppress: 26472)
              throw CParseException{PARSE_EXCEPTION_REASON::ENCODED_UNICODE_CHARACTER_TOO_SHORT, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

            //Parse out the 4 digits after the "u"
            currentChar = L'\0';
            for (int i=0; i<4; i++)
            {
#pragma warning(suppress: 26481)
              ++pszCurrentChar;
              currentChar <<= 4;
#pragma warning(suppress: 26489)
              if ((*pszCurrentChar >= L'0') && (*pszCurrentChar <= L'9'))
                currentChar |= (*pszCurrentChar - L'0');
#pragma warning(suppress: 26489)
              else if ((*pszCurrentChar >= L'A') && (*pszCurrentChar <= L'F'))
                currentChar |= (10 + (*pszCurrentChar - L'A'));
#pragma warning(suppress: 26489)
              else if ((*pszCurrentChar >= L'a') && (*pszCurrentChar <= L'f'))
                currentChar |= (10 + (*pszCurrentChar - L'a'));
              else
#pragma warning(suppress: 26472)
                throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ENCODED_UNICODE_CHARACTER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
            }
            break;
          }
          default:
          {
            if (nFlags & PARSE_FLAGS::PARSE_JSON5)
            {
              //Skip over escaped line terminators
              if (CurrentCharIsEOL(*pszCurrentChar))
              {
#pragma warning(suppress: 26481)
                if ((*pszCurrentChar == L'\r') && (*(pszCurrentChar + 1) == L'\n'))
#pragma warning(suppress: 26481 26487)
                  ++pszCurrentChar;
#pragma warning(suppress: 26481)
                ++pszCurrentChar;
                continue;
              }
              else
              {
                switch (*pszCurrentChar)
                {
                  case L'v':
                  {
                    currentChar = L'\v';
                    break;
                  }
                  case L'0':
                  {
                    currentChar = L'\0';
                    break;
                  }
                  case L'x':
                  {
                    //We need 2 digits after the "x"
#pragma warning(suppress: 26486)
                    if (wcslen(pszCurrentChar) < 3)
#pragma warning(suppress: 26472)
                      throw CParseException{PARSE_EXCEPTION_REASON::ENCODED_UNICODE_CHARACTER_TOO_SHORT, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

                    //Parse out the 2 digits after the "x"
                    currentChar = L'\0';
                    for (int i=0; i<2; i++)
                    {
#pragma warning(suppress: 26481)
                      ++pszCurrentChar;
                      currentChar <<= 4;
#pragma warning(suppress: 26489)
                      if ((*pszCurrentChar >= L'0') && (*pszCurrentChar <= L'9'))
                        currentChar |= (*pszCurrentChar - L'0');
#pragma warning(suppress: 26489)
                      else if ((*pszCurrentChar >= L'A') && (*pszCurrentChar <= L'F'))
                        currentChar |= (10 + (*pszCurrentChar - L'A'));
#pragma warning(suppress: 26489)
                      else if ((*pszCurrentChar >= L'a') && (*pszCurrentChar <= L'f'))
                        currentChar |= (10 + (*pszCurrentChar - L'a'));
                      else
#pragma warning(suppress: 26472)
                        throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ENCODED_UNICODE_CHARACTER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
                    }
                    break;
                  }
                  default:
                  {
                    //decimal digits 1 through 9 are not allowed to be escaped in JSON5
                    if ((*pszCurrentChar >= L'1') && (*pszCurrentChar <= L'9'))
#pragma warning(suppress: 26472)
                      throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON5_IDENTIFIER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
                    currentChar = *pszCurrentChar;
                    break;
                  }
                }
              }
            }
            else
#pragma warning(suppress: 26472)
              throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON5_IDENTIFIER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
            break;
          }
        }
      }
      else if (CurrentCharIsWhitespace(currentChar, nFlags) || (currentChar == L':')) //We have hit the end of an identifier if the current character is whitespace or a ":"
        return sReturn;
      else if (currentChar < L' ') //A character which is not allowed
#pragma warning(suppress: 26472)
        throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON5_IDENTIFIER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};

      sReturn += currentChar;
#pragma warning(suppress: 26481)
      ++pszCurrentChar;
    }

#pragma warning(suppress: 26472)
    throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_JSON5_IDENTIFIER, static_cast<std::size_t>(pszCurrentChar - pszJSON)};
  }

  static std::wstring IndentString(_In_ std::uint32_t nFlags, _In_ std::wstring::size_type nIndentLevel)
  {
    if (nFlags & FOUR_SPACES)
      return {nIndentLevel * 4, L' ', std::wstring::allocator_type{}};
    else if (nFlags & THREE_SPACES)
      return {nIndentLevel * 3, L' ', std::wstring::allocator_type{}};
    else if (nFlags & TWO_SPACES)
      return {nIndentLevel * 2, L' ', std::wstring::allocator_type{}};
    else if (nFlags & ONE_SPACE)
      return {nIndentLevel, L' ', std::wstring::allocator_type{}};
    else if (nFlags & TABS)
      return {nIndentLevel, L'\t', std::wstring::allocator_type{}};
    else
      return {};
  }

  [[nodiscard]] std::wstring Encode(_In_ std::uint32_t nFlags, _In_ std::size_t nIndentLevel) const
  {
    //What will be the return value from this method
    std::wstring sString;

    const std::size_t nNextIndentLevel{nIndentLevel + 1};
    const std::wstring sIndent1{IndentString(nFlags, nIndentLevel)};
    const std::wstring sIndent2{IndentString(nFlags, nNextIndentLevel)};

    switch (m_Type)
    {
      case TYPE::NULL_VALUE:
      {
        HandleBeforeValue(sString, nFlags);
        sString += L"null";
        HandleAfterValue(sString, nFlags);
        break;
      }
      case TYPE::BOOLEAN:
      {
        HandleBeforeValue(sString, nFlags);
        if (AsBoolean())
          sString += L"true";
        else
          sString += L"false";
        HandleAfterValue(sString, nFlags);
        break;
      }
      case TYPE::NUMBER:
      {
        HandleBeforeValue(sString, nFlags);
        const double fNumber{AsNumber()};
        if (nFlags & ENCODE_FLAGS::ENCODE_JSON5)
        {
          //Handle the three special cases for JSON5 numbers
          if (std::isnan(fNumber))
            sString += L"NaN";
          else if (fNumber == std::numeric_limits<double>::infinity())
            sString += L"Infinity";
          else if (fNumber == -std::numeric_limits<double>::infinity())
            sString += L"-Infinity";
          else
          {
            std::wstringstream ss;
            ss.precision(15);
            ss << fNumber;
            sString += ss.str();
          }
        }
        else
        {
          if (std::isinf(fNumber) || std::isnan(fNumber)) //Treat infinity and NaN as a JSON null
            sString += L"null";
          else
          {
            std::wstringstream ss;
            ss.precision(15);
            ss << fNumber;
            sString += ss.str();
          }
        }
        HandleAfterValue(sString, nFlags);
        break;
      }
      case TYPE::STRING:
      {
        sString = EncodeString(AsString(), nFlags);
        break;
      }
      case TYPE::OBJECT:
      {
        HandleBeforeValue(sString, nFlags);
        if (nIndentLevel && (nFlags & INDENT_SUB_OBJECTS))
        {
          sString += L"{\n";
          sString += sIndent2;
        }
        else
          sString += L"{";
        const Object& object{AsObject()};
        auto iter{object.cbegin()};
        while (iter != object.cend())
        {
          sString += EncodeString(iter->first, nFlags);
          if (nFlags & WHITESPACE_BETWEEN_OBJECT_NAME_AND_VALUES)
          {
            if (nFlags & TABS)
              sString += L":\t";
            else if (nFlags & FOUR_SPACES)
              sString += L":    ";
            else if (nFlags & THREE_SPACES)
              sString += L":   ";
            else if (nFlags & TWO_SPACES)
              sString += L":  ";
            else if (nFlags & ONE_SPACE)
              sString += L": ";
            else
              sString += L":";
          }
          else
            sString += L":";
          sString += iter->second.Encode(nFlags, nNextIndentLevel);
          if (++iter != object.cend())
            sString += L",";
        }
        if (nIndentLevel && (nFlags & INDENT_SUB_OBJECTS))
        {
          sString += L"\n" ;
          sString += sIndent1;
          sString += L"}";
        }
        else
          sString += L"}";
        HandleAfterValue(sString, nFlags);
        break;
      }
      case TYPE::ARRAY:
      {
        HandleBeforeValue(sString, nFlags);
        if (nIndentLevel && (nFlags & INDENT_SUB_ARRAYS))
        {
          sString += L"[\n";
          sString += sIndent2;
        }
        else
          sString += L"[";
        const Array& array{AsArray()};
        auto iter{array.cbegin()};
        while (iter != array.cend())
        {
          sString += iter->Encode(nFlags, nNextIndentLevel);
          if (++iter != array.cend())
            sString += L",";
        }
        if (nIndentLevel && (nFlags & INDENT_SUB_ARRAYS))
        {
          sString += L"\n";
          sString += sIndent1;
          sString += L"]";
        }
        else
          sString += L"]";
        HandleAfterValue(sString, nFlags);
        break;
      }
      default:
      {
      #ifdef _DEBUG
        assert(false);
      #endif //#ifdef _DEBUG
        break;
      }
    }

    return sString;
  }

#pragma warning(suppress: 26429)
  static int ExtractJSONPointerInt(_In_z_ const wchar_t* pszJSON)
  {
    //We need at least one digit for an int
    if ((*pszJSON < L'0') || (*pszJSON > L'9'))
      throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_NUMBER, 0};

    //What will be the return value from this function
    int nNumber{0};

#pragma warning(suppress: 26489)
    while ((*pszJSON != L'\0') && (*pszJSON >= L'0') && (*pszJSON <= L'9'))
    {
#pragma warning(suppress: 26489)
      nNumber = (nNumber * 10) + (*pszJSON - L'0');

      //Prepare for the next time around
#pragma warning(suppress: 26481)
      ++pszJSON;
    }

    return nNumber;
  }

#pragma warning(suppress: 26429)
  static std::wstring ExtractJSONPointerString(_In_z_ const wchar_t* pszJSONPointer, _Inout_ const wchar_t*& pszCurrentChar)
  {
    //What will be the return value from this method
    std::wstring sReturn{L""};

#pragma warning(suppress: 26489)
    while (*pszCurrentChar != L'\0')
    {
#pragma warning(suppress: 26489)
      wchar_t currentChar{*pszCurrentChar};

      if (currentChar == L'\\') //The string has been escaped
      {
#pragma warning(suppress: 26481)
        ++pszCurrentChar;

        switch (*pszCurrentChar)
        {
          case L'"':
          {
            currentChar = L'"';
            break;
          }
          case L'\\':
          {
            currentChar = L'\\';
            break;
          }
          case L'/':
          {
            currentChar = L'/';
            break;
          }
          case L'b':
          {
            currentChar = L'\b';
            break;
          }
          case L'f':
          {
            currentChar = L'\f';
            break;
          }
          case L'n':
          {
            currentChar = L'\n';
            break;
          }
          case L'r':
          {
            currentChar = L'\r';
            break;
          }
          case L't':
          {
            currentChar = L'\t';
            break;
          }
          case L'u':
          {
            //We need 4 digits after the "u"
#pragma warning(suppress: 26486)
            if (wcslen(pszCurrentChar) < 5)
#pragma warning(suppress: 26472)
              throw CParseException{PARSE_EXCEPTION_REASON::ENCODED_UNICODE_CHARACTER_TOO_SHORT, static_cast<std::size_t>(pszCurrentChar - pszJSONPointer)};

            //Parse out the 4 digits after the "u"
            currentChar = L'\0';
            for (int i=0; i<4; i++)
            {
#pragma warning(suppress: 26481)
              ++pszCurrentChar;
              currentChar <<= 4;
#pragma warning(suppress: 26489)
              if ((*pszCurrentChar >= L'0') && (*pszCurrentChar <= L'9'))
#pragma warning(suppress: 26489)
                currentChar |= (*pszCurrentChar - L'0');
#pragma warning(suppress: 26489)
              else if ((*pszCurrentChar >= L'A') && (*pszCurrentChar <= L'F'))
                currentChar |= (10 + (*pszCurrentChar - 'A'));
#pragma warning(suppress: 26489)
              else if ((*pszCurrentChar >= L'a') && (*pszCurrentChar <= L'f'))
                currentChar |= (10 + (*pszCurrentChar - L'a'));
              else
#pragma warning(suppress: 26472)
                throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_ENCODED_UNICODE_CHARACTER, static_cast<std::size_t>(pszCurrentChar - pszJSONPointer)};
            }
            break;
          }
          default:
          {
#pragma warning(suppress: 26472)
            throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_STRING, static_cast<std::size_t>(pszCurrentChar - pszJSONPointer)};
            break;
          }
        }
      }
      else if (currentChar == L'/') //The start of the next token
      {
        DecodeJSONPointerToken(sReturn);
        return sReturn;
      }
      else if (currentChar < L' ') //A character which is not allowed
#pragma warning(suppress: 26472)
        throw CParseException{PARSE_EXCEPTION_REASON::UNEXPECTED_CHARACTER_IN_STRING, static_cast<std::size_t>(pszCurrentChar - pszJSONPointer)};

      sReturn += currentChar;
#pragma warning(suppress: 26481 26487)
      ++pszCurrentChar;
    }

    DecodeJSONPointerToken(sReturn);
    return sReturn;
  }

  static void DecodeJSONPointerToken(_Inout_ std::wstring& sToken)
  {
    auto nFind{sToken.find(L"~1")};
    while (nFind != std::wstring::npos)
    {
      sToken.replace(nFind, 2, L"/");
      nFind = sToken.find(L"~1");
    }
    nFind = sToken.find(L"~0");
    while (nFind != std::wstring::npos)
    {
      sToken.replace(nFind, 2, L"~");
      nFind = sToken.find(L"~0");
    }
  }
};


}; //namespace JSONPP

#endif //#ifndef __JSONPP_H__
